package tes;

import java.util.*;
import java.io.*;

/**
 * RuleBaseSystem
 * 
 */
public class RuleBaseSystem {
	static RuleBase rb;

	public static void main(String args[]) {
		rb = new RuleBase();
		rb.forwardChain();
	}
}

/**
 * ワーキングメモリを表すクラス．
 * 
 * 
 */
class WorkingMemory {
	ArrayList<String> assertions;

	WorkingMemory() {
		assertions = new ArrayList<String>();
	}

	/**
	 * マッチするアサーションに対するバインディング情報を返す （再帰的）
	 * 
	 * @param 前件を示す
	 *            ArrayList
	 * @return matchableメソッドから得たバインディング情報が入っている ArrayList
	 */
	public ArrayList<HashMap<String, String>> matchingAssertions(
			ArrayList<String> theAntecedents) {
		ArrayList<HashMap<String, String>> bindings = new ArrayList<HashMap<String, String>>();
		return matchable(theAntecedents, 0, bindings);
	}
	/**
	 * マッチするアサーションに対するバインディング情報を返す （再帰的）
	 * 
	 * @param 前件を示す
	 *            ArrayList
	 * @param 照合を行った回数
	 * @param バインディング情報が入っている(途中経過)
	 *            ArrayList
	 * @return バインディング情報が入っている ArrayList
	 */
	private ArrayList<HashMap<String, String>> matchable(
			ArrayList<String> theAntecedents, int n,
			ArrayList<HashMap<String, String>> bindings) {
		// 前件の数だけ繰り返す
		if (n == theAntecedents.size()) {
			return bindings;
		}
		// １つ目
		else if (n == 0) {
			boolean success = false;
			// わかっている知識の数だけループ
			for (int i = 0; i < assertions.size(); i++) {
				// バインディング情報を保持するハッシュマップ
				HashMap<String, String> binding = new HashMap<String, String>();
				// マッチングに成功
				if ((new Matcher()).matching(theAntecedents.get(n),
						assertions.get(i), binding)) {
					// バインディング情報をハッシュマップに追加
					bindings.add(binding);
					success = true;
				}
			}
			if (success) {
				return matchable(theAntecedents, n + 1, bindings);
			} else {
				return null;
			}
		}
		// ２つ目以降
		else {
			boolean success = false;
			// バインディング情報を保持するハッシュマップ
			ArrayList<HashMap<String, String>> newBindings = new ArrayList<HashMap<String, String>>();
			// 得られたバインディング情報の数だけループ
			for (int i = 0; i < bindings.size(); i++) {
				// わかっている知識の数だけループ
				for (int j = 0; j < assertions.size(); j++) {
					// マッチングに成功
					if ((new Matcher()).matching(theAntecedents.get(n),
							assertions.get(j), bindings.get(i))) {
						// バインディング情報をハッシュマップに追加
						newBindings.add(bindings.get(i));
						success = true;
					}
				}
			}
			if (success) {
				return matchable(theAntecedents, n + 1, newBindings);
			} else {
				return null;
			}
		}
	}

	/**
	 * アサーションをワーキングメモリに加える．
	 * 
	 * @param アサーションを表す
	 *            String
	 */
	public void addAssertion(String theAssertion) {
		System.out.println("ADD:" + theAssertion);
		assertions.add(theAssertion);
	}

	/**
	 * 指定されたアサーションがすでに含まれているかどうかを調べる．
	 * 
	 * @param アサーションを表す
	 *            String
	 * @return 含まれていれば true，含まれていなければ false
	 */
	public boolean contains(String theAssertion) {
		return assertions.contains(theAssertion);
	}

	/**
	 * ワーキングメモリの情報をストリングとして返す．
	 * 
	 * @return ワーキングメモリの情報を表す String
	 */
	public String toString() {
		return assertions.toString();
	}
}

/**
 * ルールベースを表すクラス．
 * 
 * 
 */
class RuleBase {
	String fileName;
	FileReader f;
	StreamTokenizer st;
	WorkingMemory wm;
	ArrayList<Rule> rules;

	RuleBase() {
		fileName = "CarShop.data";
		wm = new WorkingMemory();
		wm.addAssertion("my-car is inexpensive");
		wm.addAssertion("my-car has a VTEC engine");
		wm.addAssertion("my-car is stylish");
		wm.addAssertion("my-car has several color models");
		wm.addAssertion("my-car has several seats");
		wm.addAssertion("my-car is a wagon");
		rules = new ArrayList<Rule>();
		loadRules(fileName);
	}

	/**
	 * 前向き推論を行うためのメソッド
	 * 
	 */
	public void forwardChain() {
		boolean newAssertionCreated;
		// 新しいアサーションが生成されなくなるまで続ける．
		do {
			// 新しいアサーションが生成されたかどうかを保存する変数
			newAssertionCreated = false;
			// ルールの数だけループ
			for (int i = 0; i < rules.size(); i++) {
				// ルールを取り出す
				Rule aRule = rules.get(i);
				// 取り出されたルールの表示
				System.out.println("apply rule:" + aRule.getName());
				// ルールの前件を表示
				ArrayList<String> antecedents = aRule.getAntecedents();
				// ルールの後件を表示
				String consequent = aRule.getConsequent();
				// バインディング情報の取得
				ArrayList<HashMap<String, String>> bindings = wm
						.matchingAssertions(antecedents);

				if (bindings != null) {
					for (int j = 0; j < bindings.size(); j++) {
						// 後件をインスタンシエーション(変数にバインディング情報を当てはめる)
						String newAssertion = instantiate((String) consequent,
								bindings.get(j));
						// ワーキングメモリーになければ成功
						if (!wm.contains(newAssertion)) {
							System.out.println("Success: " + newAssertion);
							// 持っている知識に追加
							wm.addAssertion(newAssertion);
							newAssertionCreated = true;
						}
					}
				}

			}
			System.out.println("Working Memory" + wm);
		} while (newAssertionCreated);
		System.out.println("No rule produces a new assertion");
	}

	private String instantiate(String thePattern, HashMap theBindings) {
		String result = new String();
		StringTokenizer st = new StringTokenizer(thePattern);
		for (int i = 0; i < st.countTokens();) {
			String tmp = st.nextToken();
			if (var(tmp)) {
				result = result + " " + (String) theBindings.get(tmp);
			} else {
				result = result + " " + tmp;
			}
		}
		return result.trim();
	}

	private boolean var(String str1) {
		// 先頭が ? なら変数
		return str1.startsWith("?");
	}

	private void loadRules(String theFileName) {
		String line;
		try {
			int token;
			f = new FileReader("src/tes/" + theFileName);
			st = new StreamTokenizer(f);
			while ((token = st.nextToken()) != StreamTokenizer.TT_EOF) {
				switch (token) {
				case StreamTokenizer.TT_WORD:
					String name = null;
					ArrayList<String> antecedents = null;
					String consequent = null;
					if ("rule".equals(st.sval)) {
						if (st.nextToken() == '"') {
							name = st.sval;
							st.nextToken();
							if ("if".equals(st.sval)) {
								antecedents = new ArrayList<String>();
								st.nextToken();
								while (!"then".equals(st.sval)) {
									antecedents.add(st.sval);
									st.nextToken();
								}
								if ("then".equals(st.sval)) {
									st.nextToken();
									consequent = st.sval;
								}
							}
						}
					}
					rules.add(new Rule(name, antecedents, consequent));
					break;
				default:
					System.out.println(token);
					break;
				}
			}
		} catch (Exception e) {
			System.out.println(e);
		}
		for (int i = 0; i < rules.size(); i++) {
			System.out.println(rules.get(i).toString());
		}
	}
}

/**
 * ルールを表すクラス．
 * 
 * 
 */
class Rule {
	String name;
	ArrayList<String> antecedents;
	String consequent;

	Rule(String theName, ArrayList<String> theAntecedents, String theConsequent) {
		this.name = theName;
		this.antecedents = theAntecedents;
		this.consequent = theConsequent;
	}

	/**
	 * ルールの名前を返す．
	 * 
	 * @return 名前を表す String
	 */
	public String getName() {
		return name;
	}

	/**
	 * ルールをString形式で返す
	 * 
	 * @return ルールを整形したString
	 */
	public String toString() {
		return name + " " + antecedents.toString() + "->" + consequent;
	}

	/**
	 * ルールの前件を返す．
	 * 
	 * @return 前件を表す ArrayList
	 */
	public ArrayList<String> getAntecedents() {
		return antecedents;
	}

	/**
	 * ルールの後件を返す．
	 * 
	 * @return 後件を表す String
	 */
	public String getConsequent() {
		return consequent;
	}
}

class Matcher {
	StringTokenizer st1;
	StringTokenizer st2;
	HashMap<String, String> vars;

	Matcher() {
		vars = new HashMap<String, String>();
	}

	public boolean matching(String string1, String string2,
			HashMap<String, String> bindings) {
		this.vars = bindings;
		return matching(string1, string2);
	}

	public boolean matching(String string1, String string2) {
		// System.out.println(string1);
		// System.out.println(string2);

		// 同じなら成功
		if (string1.equals(string2))
			return true;

		// 各々トークンに分ける
		st1 = new StringTokenizer(string1);
		st2 = new StringTokenizer(string2);

		// 数が異なったら失敗
		if (st1.countTokens() != st2.countTokens())
			return false;

		// 定数同士
		for (int i = 0; i < st1.countTokens();) {
			if (!tokenMatching(st1.nextToken(), st2.nextToken())) {
				// トークンが一つでもマッチングに失敗したら失敗
				return false;
			}
		}

		// 最後まで O.K. なら成功
		return true;
	}

	boolean tokenMatching(String token1, String token2) {
		// System.out.println(token1+"<->"+token2);
		if (token1.equals(token2))
			return true;
		if (var(token1) && !var(token2))
			return varMatching(token1, token2);
		if (!var(token1) && var(token2))
			return varMatching(token2, token1);
		return false;
	}

	boolean varMatching(String vartoken, String token) {
		if (vars.containsKey(vartoken)) {
			if (token.equals(vars.get(vartoken))) {
				return true;
			} else {
				return false;
			}
		} else {
			vars.put(vartoken, token);
		}
		return true;
	}

	boolean var(String str1) {
		// 先頭が ? なら変数
		return str1.startsWith("?");
	}
}
